@@ -569,20 +569,24 @@ class BaseExperiment:
         y_train_predict_proba = self.model.predict_proba(self.X_train)
         y_val_predict_proba = self.model.predict_proba(self.X_val)
         y_test_predict_proba = self.model.predict_proba(self.X_test)
-        if self.n_classes==2:
+        if self.n_classes is None:
+            y_train_predict_proba = y_val_predict_proba = y_test_predict_proba = None
+
+        elif self.n_classes==2:
             y_train_predict_proba = y_train_predict_proba[:,1]
             y_val_predict_proba = y_val_predict_proba[:,1]
             y_test_predict_proba = y_test_predict_proba[:,1]
 
-        train_auc = roc_auc_score(y_true=self.y_train, y_score=y_train_predict_proba)
-        val_auc = roc_auc_score(y_true=self.y_val, y_score=y_val_predict_proba)   
-        test_auc = roc_auc_score(y_true=self.y_test, y_score=y_test_predict_proba)
-        self.logger.add_scalar('val/acc', val_acc)
-        self.logger.add_scalar('train/acc', train_acc)
-        self.logger.add_scalar('test/acc', test_acc)
-        self.logger.add_scalar('val/auc', val_auc)
-        self.logger.add_scalar('train/auc', train_auc)
-        self.logger.add_scalar('test/auc', test_auc)
+        if self.n_classes:
+            train_auc = roc_auc_score(y_true=self.y_train, y_score=y_train_predict_proba)
+            val_auc = roc_auc_score(y_true=self.y_val, y_score=y_val_predict_proba)   
+            test_auc = roc_auc_score(y_true=self.y_test, y_score=y_test_predict_proba)
+            self.logger.add_scalar('val/acc', val_acc)
+            self.logger.add_scalar('train/acc', train_acc)
+            self.logger.add_scalar('test/acc', test_acc)
+            self.logger.add_scalar('val/auc', val_auc)
+            self.logger.add_scalar('train/auc', train_auc)
+            self.logger.add_scalar('test/auc', test_auc)
         if sample_weight:
             self.logger.add_scalar('sample_weight', sample_weight)
         self.logger.flush()
