@@ -1,3 +1,4 @@
+from ast import Assert
 from pickle import BINBYTES
 from random import sample
 import torch
@@ -9,7 +10,7 @@ from astrape.exceptions.exceptions import *
 from astrape.models.model_buildingblocks import Down, DoubleConv, Up
 import math
 import numpy as np
-
+from astrape.constants import *
 
 
 class _MLP(nn.Module):
@@ -87,7 +88,9 @@ class _MLP(nn.Module):
         self.activation_type = activation_type
         self.negative_slope = negative_slope
         if self.n_layers == 1:
-            if self.n_classes > 2:
+            if self.n_classes is None:
+                layers.append(nn.Linear(self.dims, 1))
+            elif self.n_classes > 2:
                 layers.append(nn.Linear(self.dims, self.n_classes))
             else:
                 layers.append(nn.Linear(self.dims, 1))
@@ -128,8 +131,9 @@ class _MLP(nn.Module):
                     raise ValueError(f"Wrong name {activation_type} for activation function.")
                 layers.append(nn.Dropout(self.dropout_p))
 
-            
-            if self.n_classes > 2:
+            if self.n_classes is None:
+                layers.append(nn.Linear(self.n_hidden_units, 1))
+            elif self.n_classes > 2:
                 fc3 = nn.Linear(self.n_hidden_units, self.n_classes)
                 layers.append(fc3)
             else: # including the case when self.n_classes is None
@@ -329,42 +333,19 @@ class _ContractingMLP(nn.Module):
         
         
         layers = []
+        if self.n_classes:
+            if int(self.current_units * self.contraction_ratio) <= self.n_classes:
+                if self.n_classes > 2:
+                    layers.append(nn.Linear(self.dims, self.n_classes))
+                else:
+                    layers.append(nn.Linear(self.dims, 1))
 
-        if int(self.current_units * self.contraction_ratio) <= self.n_classes:
-            if self.n_classes > 2:
-                layers.append(nn.Linear(self.dims, self.n_classes))
-            else:
-                layers.append(nn.Linear(self.dims, 1))
-
-        else : #Deep  Neural Network
-            self.current_units = int(self.current_units * self.contraction_ratio)
-            fc1 = nn.Linear(self.dims, self.current_units)
-            layers.append(fc1)
-            if self.bn:
-                layers.append(nn.BatchNorm1d(num_features=self.current_units))
-            if activation_type == "tanh":
-                layers.append(nn.Tanh())
-            elif activation_type == "relu":
-                layers.append(nn.ReLU())
-            elif activation_type == "leaky_relu":
-                if not negative_slope:
-                    negative_slope = 1e-2
-                layers.append(nn.LeakyReLU(negative_slope=negative_slope))
-            else:
-                raise ValueError(f"Wrong name {activation_type} for activation function.")
-            
-            layers.append(nn.Dropout(self.dropout_p))
-
-            
-            while int(self.current_units * self.contraction_ratio) >= self.n_classes:
-
-                
-                fc2 = nn.Linear(self.current_units,  int(self.current_units * self.contraction_ratio))
+            else : #Deep  Neural Network
                 self.current_units = int(self.current_units * self.contraction_ratio)
-                layers.append(fc2)
+                fc1 = nn.Linear(self.dims, self.current_units)
+                layers.append(fc1)
                 if self.bn:
                     layers.append(nn.BatchNorm1d(num_features=self.current_units))
-
                 if activation_type == "tanh":
                     layers.append(nn.Tanh())
                 elif activation_type == "relu":
@@ -377,22 +358,98 @@ class _ContractingMLP(nn.Module):
                     raise ValueError(f"Wrong name {activation_type} for activation function.")
                 
                 layers.append(nn.Dropout(self.dropout_p))
+
                 
-            if self.n_classes > 2:
-        
-                fc3 = nn.Linear(self.current_units, self.n_classes)
-                layers.append(fc3)
+                while int(self.current_units * self.contraction_ratio) >= self.n_classes:
+
+                    
+                    fc2 = nn.Linear(self.current_units,  int(self.current_units * self.contraction_ratio))
+                    self.current_units = int(self.current_units * self.contraction_ratio)
+                    layers.append(fc2)
+                    if self.bn:
+                        layers.append(nn.BatchNorm1d(num_features=self.current_units))
+
+                    if activation_type == "tanh":
+                        layers.append(nn.Tanh())
+                    elif activation_type == "relu":
+                        layers.append(nn.ReLU())
+                    elif activation_type == "leaky_relu":
+                        if not negative_slope:
+                            negative_slope = 1e-2
+                        layers.append(nn.LeakyReLU(negative_slope=negative_slope))
+                    else:
+                        raise ValueError(f"Wrong name {activation_type} for activation function.")
+                    
+                    layers.append(nn.Dropout(self.dropout_p))
+                    
+                if self.n_classes > 2:
+            
+                    fc3 = nn.Linear(self.current_units, self.n_classes)
+                    layers.append(fc3)
+                else:
+                    fc3 = nn.Linear(self.current_units, 1)
+                    layers.append(fc3)
+            
+            if self.n_classes> 2:
+                layers.append(nn.Softmax(dim=1))
             else:
+                layers.append(nn.Sigmoid())
+
+
+            self.layers = nn.Sequential(*layers)
+        else: # regression
+
+            if int(self.current_units * self.contraction_ratio) <= 1:
+                layers.append(nn.Linear(self.dims, 1))
+
+            else : #Deep  Neural Network
+                self.current_units = int(self.current_units * self.contraction_ratio)
+                fc1 = nn.Linear(self.dims, self.current_units)
+                layers.append(fc1)
+                if self.bn:
+                    layers.append(nn.BatchNorm1d(num_features=self.current_units))
+                if activation_type == "tanh":
+                    layers.append(nn.Tanh())
+                elif activation_type == "relu":
+                    layers.append(nn.ReLU())
+                elif activation_type == "leaky_relu":
+                    if not negative_slope:
+                        negative_slope = 1e-2
+                    layers.append(nn.LeakyReLU(negative_slope=negative_slope))
+                else:
+                    raise ValueError(f"Wrong name {activation_type} for activation function.")
+                
+                layers.append(nn.Dropout(self.dropout_p))
+
+                
+                while int(self.current_units * self.contraction_ratio) > 1:
+
+                    
+                    fc2 = nn.Linear(self.current_units,  int(self.current_units * self.contraction_ratio))
+                    self.current_units = int(self.current_units * self.contraction_ratio)
+                    layers.append(fc2)
+                    if self.bn:
+                        layers.append(nn.BatchNorm1d(num_features=self.current_units))
+
+                    if activation_type == "tanh":
+                        layers.append(nn.Tanh())
+                    elif activation_type == "relu":
+                        layers.append(nn.ReLU())
+                    elif activation_type == "leaky_relu":
+                        if not negative_slope:
+                            negative_slope = 1e-2
+                        layers.append(nn.LeakyReLU(negative_slope=negative_slope))
+                    else:
+                        raise ValueError(f"Wrong name {activation_type} for activation function.")
+                    
+                    layers.append(nn.Dropout(self.dropout_p))
+                    
                 fc3 = nn.Linear(self.current_units, 1)
                 layers.append(fc3)
-        
-        if self.n_classes> 2:
-            layers.append(nn.Softmax(dim=1))
-        else:
+            
+
             layers.append(nn.Sigmoid())
-
-
-        self.layers = nn.Sequential(*layers)
+            self.layers = nn.Sequential(*layers)
 
 
     def forward(self, X):
@@ -411,7 +468,7 @@ class ContractingMLP(BaseNet):
     def __init__(
         self,
         dims : Union[int, Tuple[int]],
-        n_classes : int,
+        n_classes : Optional[int],
         sample_weight : Optional[Union["torch.Tensor", "np.ndarray"]] = None,
         **hparams : Any
     ) -> None:
@@ -543,7 +600,7 @@ class _CustomMLP(nn.Module):
         weight_decay : float,
         l1_strength : float,
         activation_type : str = "relu",
-        negative_slope : Optional[float] = None,
+        negative_slope : Optional[float] = 0.2,
         n_hidden_units_list : Optional[List[int]] = None,
         sample_weight : Optional[Union["torch.Tensor", "np.ndarray"]] = None
     ) -> None:
@@ -574,7 +631,9 @@ class _CustomMLP(nn.Module):
         layers = []
 
         if self.n_hidden_units_list is None:
-            if self.n_classes > 2:
+            if self.n_classes is None:
+                layers.append(nn.Linear(self.dims, 1))
+            elif self.n_classes > 2:
                 layers.append(nn.Linear(self.dims, self.n_classes))
             else:
                 layers.append(nn.Linear(self.dims, 1))
@@ -589,7 +648,7 @@ class _CustomMLP(nn.Module):
                 layers.append(nn.ReLU())
             elif activation_type == "leaky_relu":
                 if not negative_slope:
-                    negative_slope = 1e-2
+                    negative_slope = 0.2
                 layers.append(nn.LeakyReLU(negative_slope=negative_slope))
             else:
                 raise ValueError(f"Wrong name {activation_type} for activation function.")
@@ -610,13 +669,16 @@ class _CustomMLP(nn.Module):
                 else:
                     raise ValueError(f"Wrong name {activation_type} for activation function.")
                 layers.append(nn.Dropout(self.dropout_p))
-            
-            if self.n_classes > 2:
+            if self.n_classes is None:
+                layers.append(nn.Linear(the_rest[-1], 1))
+            elif self.n_classes > 2:
                 layers.append(nn.Linear(the_rest[-1], self.n_classes))
             else:
                 layers.append(nn.Linear(the_rest[-1], 1))
         
-        if self.n_classes> 2:
+        if self.n_classes is None:
+            pass
+        elif self.n_classes > 2:
             layers.append(nn.Softmax(dim=1))
         else:
             layers.append(nn.Sigmoid())
@@ -874,17 +936,6 @@ class VGG(ConvNet):
 
     def forward(self, X):
         return self.model.forward(X)
-
-    def help(self) -> None :
-        whoami = "VGG Network\n\n"
-        helptext = "Parameters ------------------------------------------------------------------------------------------------------- \n"
-        helptext += "n_classes: int \t Number of classes in the classification task. If it is binary classification, n_classes=2\n"
-        helptext += "cfg: str\t The configuration of VGG. 'VGG11', 'VGG13', 'VGG16', and 'VGG19' are available.\n"
-        helptext += "In defining hyperparameters and optimizer_type, we recommend you to define a dictionary.\n"
-        helptext += "Define a dictionary containing :\n\t-'optimizer_type': str (default='adam')\n\t-'dropout_p' : 0 <= float < 1(default : 0)\n\t-'weight_decay' : 0<= float < 1(default=0)\n\t-'lr' : float(default=1e-4)."
-        
-        text= whoami + helptext
-        print(text)
         
 class _UNet(nn.Module):
     """
@@ -1044,7 +1095,261 @@ class UNet(ConvNet):
     def forward(self, X):
         return self.model.forward(X)
     
+
+class _Generator(nn.Module):
+    def __init__(
+        self,
+        dims : Union[int, Tuple[int]],
+        batch_size : int,
+        optimizer_type : str,
+        lr : float,
+        bn : bool,
+        sample_weight : Optional[Union["np.ndarray", "torch.Tensor"]] = None,
+        hidden_units : List[int] = [128, 512, 1024],
+        latent_dim : int = 100,
+        activation_type : str = "leaky_relu",
+        negative_slope : Optional[float] = 0.2
+    )->None:
+        super(_Generator, self).__init__()
+        self.dims = dims
+        self.batch_size = batch_size
+        self.optimizer_type = optimizer_type
+        self.lr = lr
+        self.sample_weight = sample_weight
+        self.hidden_units = hidden_units
+        self.latent_dim = latent_dim
+        self.bn = bn
+        self.activation_type = activation_type
+        self.negative_slope = negative_slope
+        
+        def block(in_units, out_units):
+            layers = [nn.Linear(in_units, out_units)]
+            if self.bn:
+                layers.append(nn.BatchNorm1d(out_units, 0.8))
+            if self.activation_type == "relu":
+                layers.append(nn.ReLU())
+            elif self.activation_type == "leaky_relu":
+                layers.append(nn.LeakyReLU(self.negative_slope))
+            elif self.activation_type == "tanh":
+                layers.append(nn.Tanh())
+            else:
+                raise AssertionError(f"Wrong type of activation type {self.activation_type}. It should be one of ['relu', 'leaky_relu', 'tanh']")
+            return layers
+        
+        layers = []
+
+        for hidden_unit in self.hidden_units:
+            layers.extend(block(self.latent_dim, hidden_unit, self.bn))
+        layers.append(nn.Linear(self.hidden_units[-1], int(np.prod(self.dims))))
+        layers.append(nn.Tanh())
+
+        self.layers = nn.Sequential(*layers) 
+
+    def forward(self, x):
+        img = self.layers(x)
+        img = img.view(img.size(0), *self.dims)
+        
+        return img
+
+class _Discriminator(nn.Module):
+    def __init__(
+        self,
+        dims : Union[int, Tuple[int]],
+        batch_size : int,
+        optimizer_type : str,
+        lr : float,
+        hidden_units : List[int] = [512, 256],
+        sample_weight : Optional[Union["np.ndarray", "torch.Tensor"]] = None,
+        activation_type : str = "leaky_relu",
+        negative_slope : Optional[float] = 0.2
+    )->None:
+        super(_Discriminator, self).__init__()
+        self.dims = dims
+        self.batch_size = batch_size
+        self.optimizer_type = optimizer_type
+        self.lr = lr
+        self.hidden_units = hidden_units
+        self.sample_weight = sample_weight
+        self.activation_type = activation_type
+        self.negative_slope = negative_slope
+
+        def block(in_units, out_units):
+            layers = [nn.Linear(in_units, out_units)]
+            if self.activation_type == "relu":
+                layers.append(nn.ReLU())
+            elif self.activation_type == "leaky_relu":
+                layers.append(nn.LeakyReLU(self.negative_slope))
+            elif self.activation_type == "tanh":
+                layers.append(nn.Tanh())
+            else:
+                raise AssertionError(f"Wrong type of activation type {self.activation_type}. It should be one of ['relu', 'leaky_relu', 'tanh']")
+            return layers
+            
+        layers = []
+
+        layers.extend(block(int(np.prod(self.dims)), self.hidden_units[0]))
+        for i in range(1, len(self.hidden_units)-1):
+            layers.extend(block(self.hidden_units[i-1],self.hidden_units[i]))
+        layers.append(nn.Linear(self.hidden_units[-1], 1))
+        layers.append(nn.Sigmoid())
+        
+        self.layers = nn.Sequential(*layers)
+
+    def forward(self, img):
+        img_flat = img.view(img.size(0), -1)
+        validity = self.layers(img_flat)
+
+        return validity
     
+
+"""
+class Generator(GenerativeNet):
+    def __init__(
+        self
+    ):
+        super(Generator, self).__init__()
+    @property
+    def criterion(self, img_batch):
+        
+
+        real = Discriminator(img_batch.view(-1, int(np.prod(self.dims))))
+        fake = Disriminator(Generator())
+        criterion = None
+        return criterion
+    
+    pass
+    def training_step(self, train_batch, batch_idx):
+        x_batch, y_batch = train_batch
+        x_batch = self.reshape_x(x_batch)
+            
+        y_batch = self.sample_weight * y_batch
+
+        Yhat = self.forward(x_batch)
+        if self.n_classes is None:
+            loss = self.criterion(Yhat.squeeze(), y_batch)
+            self.log('train/loss', loss)
+            return {'loss' : loss}
+        else:    
+            if self.n_classes > 2:
+                loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.LongTensor))
+            else:
+                loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.FloatTensor))
+            self.log('train/loss', loss)
+            self.log('train/acc', acc)
+            self.log('train/auc', auc)
+            return {'loss' : loss, 'acc' : acc, 'auc' : auc}
+
+        
+    def validation_step(self, val_batch, batch_idx):
+        x_batch, y_batch = val_batch
+        x_batch = self.reshape_x(x_batch)
+        Yhat = self.forward(x_batch)
+        if self.n_classes is None:
+            loss = self.criterion(Yhat.squeeze(), y_batch)
+            self.log('val/loss', loss)
+            return {'loss' : loss}
+        else:    
+            if self.n_classes > 2:
+                loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.LongTensor))
+            else:
+                loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.FloatTensor))
+
+            self.log('val/loss', loss)
+            self.log('val/acc', acc)
+            self.log('val/auc', auc)
+            return {'loss' : loss, 'acc' : acc, 'auc' : auc}
+
+    def predict_step(self, pred_batch, batch_idx, dataloader_idx=0):
+        x_batch, y_batch = pred_batch
+        x_batch = self.reshape_x(x_batch)
+        Yhat = self.forward(x_batch)
+        if self.n_classes is None:
+            loss = self.criterion(Yhat.squeeze(), y_batch)
+            self.log('pred/loss', loss)
+            return {'loss' : loss}
+        else:    
+            if self.n_classes > 2:
+                loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.LongTensor))
+            else:
+                loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.FloatTensor))
+            self.log('pred/loss', loss)
+            self.log('pred/acc', acc)
+            self.log('pred/auc', auc)
+            return {'loss' : loss, 'acc' : acc, 'auc' : auc}
+
+    def test_step(self, test_batch, batch_idx):
+        x_batch, y_batch = test_batch
+        x_batch = self.reshape_x(x_batch)
+        Yhat = self.forward(x_batch)
+        if self.n_classes is None:
+            loss = self.criterion(Yhat.squeeze(), y_batch)
+            self.log('test/loss', loss)
+            return {'loss' : loss}
+        else:    
+            if self.n_classes > 2:
+                loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.LongTensor))
+            else:
+                loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.FloatTensor))
+            self.log('test/loss', loss)
+            self.log('test/acc', acc)
+            self.log('test/auc', auc)
+            return {'loss' : loss, 'acc' : acc, 'auc' : auc}
+"""
+
+"""
+class Discriminator(BaseNet):
+    def __init__(
+        self,
+        dims : Union[int, Tuple[int]],
+        **hparams : Any
+    ) -> None:
+        entire_keys = [
+            'dims', 
+            'hidden_units', 
+            'optimizer_type', 
+            'batch_size',
+            'lr',
+            'sample_weight',
+            'activation_type',
+            'negative_slope'
+            ]
+        default_parameter_dict = {
+            'hidden_units' : [512, 256],
+            'activation_type' : "leaky_relu",
+            'negative_slope' : 0.2
+        }
+        
+        self.dims = dims
+        
+        # set default for model-specific hyperparameters
+        for argument, default in default_parameter_dict.items():
+            if argument not in hparams.keys(): 
+                hparams.update({argument : default})
+        # set default for model-agnostic hyperparameters
+        if 'lr' not in hparams.keys():
+            hparams.update({'lr' : DEFAULT_LR})
+        if 'optimizer_type' not in hparams.keys():
+            hparams.update({'optimizer_type' : DEFAULT_OPTIMIZER_TYPE})
+        if 'batch_size' not in hparams.keys():
+            hparams.update({'batch_size' : DEFAULT_BATCH_SIZE})
+        
+        # save information of **hparams to self.hparams.
+        self.save_hyperparameters(*hparams)
+        
+        super(Discriminator, self).__init__(
+            dims=dims, 
+            **hparams
+        )
+        # sanity check. Check if right arguments are declared.
+        for key in hparams.keys():
+            if key not in entire_keys:
+                raise ValueError(f'Wrong argument {key} is passed to the model.')
+        self.model = _Discriminator(dims=self.dims, **hparams)
+
+    def forward(self, X):
+        return self.model.forward(X)
+        
+"""            
 """
 
 TODO : Implement GoogleNet
