@@ -467,9 +467,9 @@ class BaseProject:
                 if self.n_classes:
                     for label_idx in range(self.n_classes):
                         points_dict.update({label_idx: self.X[self.y==label_idx]})
-                        plt.scatter(x=points_dict[label_idx][:,0], y=points_dict[label_idx][:,1], label=label_idx)
+                        plt.scatter(x=points_dict[label_idx][:,0], label=label_idx)
                 else:
-                    plt.scatter(x=self.X[:,0], y=self.X[:,1])
+                    plt.scatter(x=self.X[:,0])
                 plt.title('Data Distribution')
                 plt.xlabel('positions')
                 plt.legend(loc=(1.05,1.05))
